#ifndef DARRAY_H
#define DARRAY_H

#include <windows.h>

#if defined(__BORLANDC__)
#   define TRY      try
#   define EXCEPT   __except
#elif defined(_MSC_VER)
#   define TRY     _try
#   define EXCEPT  _except
#endif


class CDynamicArrayBase {
protected:
    static DWORD    dwPageSize;  // System page size
    static DWORD PageMask()     { return dwPageSize-1; }

    LPSTR    pBase;              // Pointer to base of array
    DWORD    dwSize;             // Maximum number of pages
    DWORD    dwTopCommit, dwBottomCommit;
    DWORD    dwTopRequest, dwBottomRequest;
    BOOL    fFail;

    virtual LPSTR FromBase(const DWORD) const            = 0;
    virtual LPSTR ToBase(const DWORD) const              = 0;
    virtual DWORD FromOffset(const DWORD) const          = 0;
    virtual DWORD ToOffset(const DWORD) const            = 0;
    virtual DWORD ToElements(const DWORD) const          = 0;
    virtual DWORD ToElementsCorrected(const DWORD) const = 0;
    
    virtual VOID Init(DWORD, DWORD)        { }
    virtual VOID Cleanup(DWORD, DWORD)     { }
    INT ExceptionHandler(DWORD, DWORD);

public:
    BOOL    fCommit;

    CDynamicArrayBase(DWORD);
    ~CDynamicArrayBase(void);
    
    static DWORD ToBytes(const DWORD dwPages)
        { return dwPages*dwPageSize; }
    static DWORD ToPages(const DWORD dwBytes)
        { DWORD dwPages = dwBytes/dwPageSize;
          return dwPages+( ToBytes(dwPages) < dwBytes ); }

    virtual DWORD ElementSize(void) const                = 0;
    
    BOOL Fail(void)
        {
        if ( fFail )
            { fFail = FALSE; return TRUE; }
        return FALSE;
        }
    DWORD TopCommit(void) const    { return dwTopCommit;    }
    DWORD BottomCommit(void) const { return dwBottomCommit; }
    DWORD TopRequest(void) const   { return dwTopRequest;   }
    DWORD BottomRequest(void) const{ return dwBottomRequest;}

    BOOL Commit(DWORD, DWORD);
    VOID Release(DWORD, DWORD);
    VOID Release(DWORD dwFrom = 0)
        { Release(dwFrom, Size()-1); }
    DWORD Length(void) const { return ToBytes(dwSize);      }
    DWORD Size(void) const   { return ToElements(Length()); }
    DWORD PerPage(void)const {return ToElements(dwPageSize);}
};


template<class T>
class CDynamicArray : public CDynamicArrayBase {
protected:
    LPSTR FromBase(const DWORD dwIndex) const
        { return LPSTR(DWORD((T*)pBase+dwIndex)
                    & ~PageMask()); }
    LPSTR ToBase(const DWORD dwIndex) const
        { return LPSTR((DWORD((T*)pBase+dwIndex)+sizeof(T)-1)
                    & ~PageMask()); }
    DWORD FromOffset(const DWORD dwIndex) const
        { return DWORD((T*)pBase+dwIndex) & PageMask(); }
    DWORD ToOffset(const DWORD dwIndex) const
        { return (DWORD((T*)pBase+dwIndex)+sizeof(T)-1)
                    & PageMask(); }
    DWORD ToElements(const DWORD dwBytes) const
        { return dwBytes/sizeof(T); }
    DWORD ToElementsCorrected(const DWORD dwBytes) const
        { return ToElements(dwBytes)
            + ( dwBytes%sizeof(T) > 0 ); }

public:
    T            dummy;

    CDynamicArray(DWORD dwSize)
        : CDynamicArrayBase(dwSize*sizeof(T))
        { }
    ~CDynamicArray(void)            { Release(); }

    DWORD ElementSize(void) const   { return sizeof(T); }

    T& operator()(const DWORD dwIndex)
        { return ((T*)pBase)[dwIndex]; }
    T& operator[](const DWORD);
    T& operator*(void)              { return (*this)(0); }
};

template<class T>
inline T& CDynamicArray<T>::operator[](const DWORD dwIndex)
{
    // The guarded body:
    TRY {
    // Read member bytes to force fault:
        LPSTR p = LPSTR(&(*this)(dwIndex));
        for ( int i = 0 ; i < sizeof(T) ; ++i )
            volatile char c = *(p++);
        return (*this)(dwIndex); // Return ref to member.
    }
    EXCEPT ( ExceptionHandler(
                            GetExceptionCode(), dwIndex) ) {
        fFail = TRUE;  // Mark the failure.
        return dummy;  // Return reference to dummy member.
    }
}

#endif
